#!/usr/bin/env python

import rospy
import geometry_msgs.msg
from geometry_msgs.msg import Twist
import math
import time
from std_srvs.srv import Empty
from sensor_msgs.msg import LaserScan
import tf
import numpy
import geometry_msgs.msg
from math import *
from nav_msgs.msg import Odometry
from std_msgs.msg import String
from sensor_msgs.msg import Range
import tf2_ros
import copy
from collections import defaultdict
from tf.transformations import euler_from_quaternion, quaternion_from_euler
import threading
from threading import Thread

x = 0
y = 0
roll = 0
pitch = 0
yaw = 0
scan_min = 5
detect_obstacle_rate = 30 #kinect rate 30Hz
obstacle_focus = 5
obstacleLeft = [True, 160, 0.1]
obstacleMiddleLeft = [True, 160, 0.1]
obstacleMiddleRight = [True, 160, 0.1]
obstacleRight = [True, 160, 0.1]
obstacle_in_front = True
gap = [False,0,0]
followRight = 0.8
spacing = 0.8
ultrasound_dist = 0.1
successful_measurement = False

#robot movement control
class RobotMoveControl():

    def odom_callback(self, odom_pose):
        global x
        global y
        global roll, pitch, yaw
        orientation_q = odom_pose.pose.pose.orientation
        orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
        (roll, pitch, yaw) = euler_from_quaternion(orientation_list)
        x = odom_pose.pose.pose.position.x
        y = odom_pose.pose.pose.position.y

    def move(self, speed, distance):
        global x, y, yaw
        global obstacleLeft
        global obstacleMiddleLeft
        global obstacleMiddleRight
        global obstacleRight
        global followRight
        global successful_measurement

        velocityMessage = Twist()
        x0 = x
        y0 = y


        if (speed > 0.4):
            speed = 0.4

        if (speed > 0):
            rospy.loginfo("Turtlebot is moving forward")
        else:
            rospy.loginfo("Turtlebot is moving backward")

        velocityMessage.linear.x = speed
        # all velocities except x must be zero.
        velocityMessage.linear.y = 0
        velocityMessage.linear.z = 0
        velocityMessage.angular.x = 0
        velocityMessage.angular.y = 0
        velocityMessage.angular.z = 0

        distance_moved = 0.0
        loop_rate = rospy.Rate(10)  # publishing the velocity at 10 Hz
        timestamp_recorded = False
        timestamp = 0

        while distance_moved < distance:
            if timestamp_recorded == False:
                timestamp = seconds = rospy.get_time()
                timestamp_recorded = True

            if obstacleRight[0]:
                velocityMessage.angular.z = obstacleRight[2]
            elif obstacleLeft[0]:
                velocityMessage.angular.z = -obstacleLeft[2]
            else:
                velocityMessage.angular.z = 0

            if obstacleMiddleRight[0] or obstacleMiddleLeft[0]:
                velocityMessage.linear.x = 0
                if obstacleRight[1] < followRight * obstacleLeft[1]:
                    velocityMessage.angular.z = obstacleMiddleRight[2]
                else:
                    velocityMessage.angular.z = -obstacleMiddleLeft[2]
            elif not obstacleMiddleLeft[0] and not obstacleMiddleLeft[0]:
                velocityMessage.linear.x = speed

            self.velocityPublisher.publish(velocityMessage)

            loop_rate.sleep()
            distance_moved = abs(0.5 * math.sqrt(((x - x0) ** 2) + ((y - y0) ** 2)))
            #print  distance_moved

            if not (distance_moved < distance) or rospy.get_time()-timestamp > 10*distance:
                if not (distance_moved < distance):
                    rospy.loginfo("success")
                    successful_measurement = True
                else:
                    rospy.loginfo("Didn't reach the goal - move()")
                    successful_measurement = False
                break
        #stop the robot when the distance is moved
        velocityMessage.linear.x = 0
        self.velocityPublisher.publish(velocityMessage)


    def rotate(self, angular_velocity, radians):
        global yaw
        rotateMessage = Twist()

        angle_turned = 0.0
        if radians > 0:
            while(radians > 2 * pi):
                radians -= 2 * pi
        else:
            while(radians < -2 * pi):
                radians += 2 * pi

        start_angle = yaw

        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        rotateMessage.angular.z = angular_velocity

        loop_rate = rospy.Rate(10)

        while True:
            rospy.loginfo("Turtlebot is Rotating")
            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()
            end_angle = yaw

            angle_turned = end_angle - start_angle
            if (angle_turned > radians):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def rotateInit(self):
        global yaw
        rotateMessage = Twist()
        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        loop_rate = rospy.Rate(10)
        rospy.loginfo("Rotation intialization")

        while True:
            end_angle = yaw

            if end_angle < 0:
                rotateMessage.angular.z = -end_angle
            else:
                rotateMessage.angular.z = end_angle

            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()

            if (end_angle < 0.02 and end_angle > -0.02):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def rotateToAngle(self, angle):
        global yaw
        rotateMessage = Twist()
        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        loop_rate = rospy.Rate(10)
        rospy.loginfo("Rotating to angle %f rad" %angle)

        while True:
            end_angle = yaw

            if end_angle < 0 + angle:
                rotateMessage.angular.z = -end_angle + angle
            else:
                rotateMessage.angular.z = end_angle + angle

            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()

            if (end_angle < 0.02 + angle and end_angle > -0.02 + angle):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def goToGoal(self, x2, y2):
        global x, y, obstacle_in_front, successful_measurement

        dx = x2 - round(x, 2)
        dy = y2 - round(y, 2)
        theta = math.atan2(dy, dx)
        distance = sqrt(dx * dx + dy * dy)
        self.rotateToAngle(theta)
        if not obstacle_in_front:
            self.move(2, distance)
        else:
            print("Not reachable! go_to_goal()")
            successful_measurement = False

    def __init__(self):

        self.turtlebot_odom_pose = Odometry()
        self.velocityPublisher = rospy.Publisher('/cmd_vel_mux/input/teleop', Twist, queue_size=10)
        rospy.Subscriber("odom", Odometry, self.odom_callback)
        # 2 HZ
        r = rospy.Rate(20)
        r.sleep()




# grid class serves either for collecting fingerprint samples
# or for very rough map
class Grid:
    g = defaultdict(dict)

    #extreme vlues of the grid
    minX, maxX, minY, maxY = 0, 0, 0, 0

    def __getitem__(self, item):
        return self.g.get(item)

    def printGrid(self):
        for x in self.g:
            print x

    def isSet(self, x, y):
        if tuple([x,y]) in self.g:
            return True


    def addPoint(self, x, y, sample):
        #extreme values check
        if x > self.maxX:
            self.maxX = x
        elif x < self.minX:
            self.mminX = x

        if y > self.maxY:
            self.maxY = y
        elif x < self.minY:
            self.mminY = y

        self.g[tuple([x, y])] = sample


# transform coordinates from /odom to /base_footprint
# yaw extraction - quaternion to euler
# tf didn't work with real Turtlebot 2'
def get_coordinates():
    #declare tf transform
    listener = tf.TransformListener()

    try:
        listener.waitForTransform("/base_footprint", "/odom", rospy.Time(), rospy.Duration(10.0) )
        (trans,rot) = listener.lookupTransform('/base_footprint', '/odom', rospy.Time(0))
    except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException):
        rospy.Duration(1.0)

    # 2D space => pose defined by x, y and yaw
    x = trans[0]
    y = trans[1]
    euler = tf.transformations.euler_from_quaternion(rot)
    yaw = euler[2]
    pose = [x, y, yaw]
    return pose

# detect obstacle at rate - combines ultrasound detection and laser scan detection
def detect_obstacle_at_rate(rate):
    global obstacle_in_front
    global obstacleMiddleLeft
    global obstacleMiddleRight
    global ultrasound_dist
    global spacing

    rate = rospy.Rate(rate)

    while not rospy.is_shutdown():
        if obstacleMiddleLeft[0] or obstacleMiddleRight[0] or ultrasound_dist < spacing:
            obstacle_in_front = True
        else:
            obstacle_in_front = False
        rate.sleep()




# recieving ultrasound distance data
def ultrasound_callback(ultra_sound_data):
    global ultrasound_dist
    ultrasound_dist = ultra_sound_data.range




# analysing data from a scanner
# recieving scan data
def scan_callback(scan_data):
    global scan_min
    #Find minimum range
    min_value, min_index = min_in_focus(scan_data.ranges, 0, 160)
    max_value, max_index = max_range_index(scan_data.ranges)

    scan_min = max_value
    detectObstacle(scan_data.ranges)


def field_of_view(scan_data):
    return (scan_data.angle_max - scan_data.angle_min) * 180.0 / 3.14


#find the max range and its index
def min_range_index(ranges):
    ranges = [x for x in ranges if not math.isnan(x)]
    return (min(ranges), ranges.index(min(ranges)))


#find the max range
def max_range_index(ranges):
    ranges = [x for x in ranges if not math.isnan(x)]
    return (max(ranges), ranges.index(max(ranges)))


def average_between_indices(ranges, i, j):
    ranges = [x for x in ranges if not math.isnan(x)]
    slice_of_array = ranges[i: j + 1]
    return (sum(slice_of_array) / float(len(slice_of_array)))


def min_in_focus(ranges, i, j):
    ranges = [x for x in ranges if not math.isnan(x)]
    slice_of_array = ranges[i: j + 1]
    return (min(slice_of_array), ranges.index(min(slice_of_array)))


def max_in_focus(ranges, i, j):
    ranges = [x for x in ranges if not math.isnan(x)]
    slice_of_array = ranges[i: j + 1]
    return (max(slice_of_array), ranges.index(max(slice_of_array)))


def detectObstacle(ranges):
    global obstacleLeft
    global obstacleMiddleLeft
    global obstacleMiddleRight
    global obstacleRight
    global spacing
    global gap
    min_value, min_index = min_in_focus(ranges, 0, 160)
    min_value1, min_index1 = min_in_focus(ranges, 161, 320)
    min_value2, min_index2 = min_in_focus(ranges, 321, 480)
    min_value3, min_index3 = min_in_focus(ranges, 481, len(ranges))
    min_index = min_index + 1
    min_index1 = min_index1 - 160
    min_index2 = -min_index2 + 481
    min_index3 = -min_index3 + 641

    if min_value < spacing:
        obstacleRight = [True, min_index, min_value]
    elif min_value >= spacing:
        obstacleRight = [False, min_index, min_value]

    if min_value1 < spacing:
        obstacleMiddleRight = [True, min_index1, min_value1]
    elif min_value1 >= spacing:
        obstacleMiddleRight = [False, min_index1, min_value1]

    if min_value2 < spacing:
        obstacleMiddleLeft = [True, min_index2, min_value2]
    elif min_value2 >= spacing:
        obstacleMiddleLeft = [False, min_index2, min_value2]

    if min_value3 < spacing:
        obstacleLeft = [True, min_index3, min_value3]
    elif min_value3 >= spacing:
        obstacleLeft = [False, min_index3, min_value3]


def measureFP(goalX, goalY):
    global obstacleMiddleLeft
    global obstacleMiddleRight
    global spacing
    global successful_measurement
    global x, y
    originX = x
    originY = y
    successful_measurement = False
    RobotMoveControl().rotateInit()
    RobotMoveControl().rotateToAngle(math.pi/2)
    if (obstacleMiddleLeft[2] < spacing or obstacleMiddleRight[2] < spacing):
        if not (goalX, goalY + spacing) in rawMap:
            rawMap.addPoint(goalX, goalY + spacing, False)
    else:
        rawMap.addPoint(goalX, goalY + spacing, True)
    rospy.sleep(0.5)
    RobotMoveControl().rotateToAngle(math.pi)
    if (obstacleMiddleLeft[2] < spacing or obstacleMiddleRight[2] < spacing):
        if not (goalX - spacing, goalY) in rawMap:
            rawMap.addPoint(goalX - spacing, goalY, False)
    else:
        rawMap.addPoint(goalX - spacing, goalY, True)
    rospy.sleep(0.5)
    RobotMoveControl().rotateToAngle(-math.pi/2)
    if (obstacleMiddleLeft[2] < spacing or obstacleMiddleRight[2] < spacing):
        if not (goalX, goalY - spacing) in rawMap:
            rawMap.addPoint(goalX, goalY - spacing, False)
    else:
        rawMap.addPoint(goalX, goalY - spacing, True)
    rospy.sleep(0.5)
    RobotMoveControl().rotateInit()
    if (obstacleMiddleLeft[2] < spacing or obstacleMiddleRight[2] < spacing):
        if not (goalX + spacing, goalY) in rawMap:
            rawMap.addPoint(goalX + spacing, goalY, False)
    else:
        rawMap.addPoint(goalX + spacing, goalY, True)
    rospy.sleep(0.5)
    RobotMoveControl().goToGoal(goalX, goalY)
    if (successful_measurement):
        grid.addPoint(goalX, goalY, "measured values")
    else:
        RobotMoveControl().goToGoal(originX, originY)


def collectFingerprints():
    global x, y, spacing
    print("approaching initial position")
    RobotMoveControl().goToGoal(x-x % spacing, y-y % spacing)
    gx = round(x, 2)
    gy = round(y, 2)

    print("initial position X = %f Y = %f" %(gx, gy))

    theStack = [(gx, gy)]

    while len(theStack) > 0:

        gx, gy = theStack.pop()
        grid.printGrid()

        if (grid.isSet(gx, gy) or rawMap[(gx + spacing, gy)] == False):
            print("Coordinates already measured or unreachable.")
            continue

        measureFP(round(gx, 2), round(gy, 2))
        if rawMap[(gx + spacing, gy)] == True or rawMap[(gx + spacing, gy)] is None:
            theStack.append((gx + spacing, gy))  # right
        if rawMap[(gx + spacing, gy + spacing)] == True or rawMap[(gx + spacing, gy + spacing)] is None:
            theStack.append((gx + spacing, gy + spacing))  # up right
        if rawMap[(gx - spacing, gy)] == True or rawMap[(gx - spacing, gy)] is None:
            theStack.append((gx - spacing, gy))  # left
        if rawMap[(gx - spacing, gy - spacing)] == True or rawMap[(gx - spacing, gy - spacing)] is None:
            theStack.append((gx - spacing, gy - spacing))  # down left
        if rawMap[(gx, gy - spacing)] == True or rawMap[(gx, gy - spacing)] is None:
            theStack.append((gx, gy - spacing))  # down
        if rawMap[(gx + spacing, gy - spacing)] == True or rawMap[(gx + spacing, gy - spacing)] is None:
            theStack.append((gx + spacing, gy - spacing))  # down right
        if rawMap[(gx, gy + spacing)] == True or rawMap[(gx, gy + spacing)] is None:
            theStack.append((gx, gy + spacing))  # up
        if rawMap[(gx - spacing, gy + spacing)] == True or rawMap[(gx - spacing, gy + spacing)] is None:
            theStack.append((gx - spacing, gy + spacing))  # up left



if __name__ == '__main__':
    try:
        rospy.init_node('robot_control')
        rospy.Subscriber("scan", LaserScan, scan_callback)
        rospy.Subscriber("ultrasound_range", Range, ultrasound_callback)
        obstacle_detection = threading.Thread(target=detect_obstacle_at_rate, name="detect_obstacle_at_rate Thread", args=[detect_obstacle_rate])
        obstacle_detection.start()

        RobotMoveControl()
        rawMap = Grid()
        grid = Grid()

        collectFingerprints()

        time.sleep(1.0)

    except rospy.ROSInterruptException:
        rospy.loginfo("node terminated.")

