#!/usr/bin/env python

import rospy
import geometry_msgs.msg
from geometry_msgs.msg import Twist
import math
import time
from std_srvs.srv import Empty
from sensor_msgs.msg import LaserScan
import tf
import numpy
import geometry_msgs.msg
from math import *
from nav_msgs.msg import Odometry
from std_msgs.msg import String
from sensor_msgs.msg import Range
from collections import defaultdict
from tf.transformations import euler_from_quaternion, quaternion_from_euler
import actionlib
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal
from math import radians, degrees
from actionlib_msgs.msg import *
from geometry_msgs.msg import Point

x = 0
y = 0
roll = 0
pitch = 0
yaw = 0

spacing = 0.8
successful_measurement = False


#robot movement control
class RobotMoveControl():

    def odom_callback(self, odom_pose):
        global x
        global y
        global roll, pitch, yaw
        orientation_q = odom_pose.pose.pose.orientation
        orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
        (roll, pitch, yaw) = euler_from_quaternion(orientation_list)
        x = odom_pose.pose.pose.position.x
        y = odom_pose.pose.pose.position.y

    def move(self, speed, distance):
        global x, y, yaw
        global obstacleLeft
        global obstacleMiddleLeft
        global obstacleMiddleRight
        global obstacleRight
        global followRight
        global successful_measurement

        velocityMessage = Twist()
        x0 = x
        y0 = y


        if (speed > 0.4):
            speed = 0.4

        if (speed > 0):
            rospy.loginfo("Turtlebot is moving forward")
        else:
            rospy.loginfo("Turtlebot is moving backward")

        velocityMessage.linear.x = speed
        # all velocities except x must be zero.
        velocityMessage.linear.y = 0
        velocityMessage.linear.z = 0
        velocityMessage.angular.x = 0
        velocityMessage.angular.y = 0
        velocityMessage.angular.z = 0

        distance_moved = 0.0
        loop_rate = rospy.Rate(10)  # publishing the velocity at 10 Hz
        timestamp_recorded = False
        timestamp = 0

        while distance_moved < distance:
            if timestamp_recorded == False:
                timestamp = seconds = rospy.get_time()
                timestamp_recorded = True

            if obstacleRight[0]:
                velocityMessage.angular.z = obstacleRight[2]
            elif obstacleLeft[0]:
                velocityMessage.angular.z = -obstacleLeft[2]
            else:
                velocityMessage.angular.z = 0

            if obstacleMiddleRight[0] or obstacleMiddleLeft[0]:
                velocityMessage.linear.x = 0
                if obstacleRight[1] < followRight * obstacleLeft[1]:
                    velocityMessage.angular.z = obstacleMiddleRight[2]
                else:
                    velocityMessage.angular.z = -obstacleMiddleLeft[2]
            elif not obstacleMiddleLeft[0] and not obstacleMiddleLeft[0]:
                velocityMessage.linear.x = speed

            self.velocityPublisher.publish(velocityMessage)

            loop_rate.sleep()
            distance_moved = abs(0.5 * math.sqrt(((x - x0) ** 2) + ((y - y0) ** 2)))
            #print  distance_moved

            if not (distance_moved < distance) or rospy.get_time()-timestamp > 10*distance:
                if not (distance_moved < distance):
                    rospy.loginfo("success")
                    successful_measurement = True
                else:
                    rospy.loginfo("Didn't reach the goal - move()")
                    successful_measurement = False
                break
        #stop the robot when the distance is moved
        velocityMessage.linear.x = 0
        self.velocityPublisher.publish(velocityMessage)

    def rotate(self, angular_velocity, radians):
        global yaw
        rotateMessage = Twist()

        angle_turned = 0.0
        if radians > 0:
            while(radians > 2 * pi):
                radians -= 2 * pi
        else:
            while(radians < -2 * pi):
                radians += 2 * pi

        start_angle = yaw

        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        rotateMessage.angular.z = angular_velocity

        loop_rate = rospy.Rate(10)

        while True:
            rospy.loginfo("Turtlebot is Rotating")
            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()
            end_angle = yaw

            angle_turned = end_angle - start_angle
            if (angle_turned > radians):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def rotateInit(self):
        global yaw
        rotateMessage = Twist()
        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        loop_rate = rospy.Rate(10)
        rospy.loginfo("Rotation intialization")

        while True:
            end_angle = yaw

            if end_angle < 0:
                rotateMessage.angular.z = -end_angle
            else:
                rotateMessage.angular.z = end_angle

            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()

            if (end_angle < 0.02 and end_angle > -0.02):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def rotateToAngle(self, angle):
        global yaw
        rotateMessage = Twist()
        rotateMessage.linear.x = rotateMessage.linear.y = 0.0
        loop_rate = rospy.Rate(10)
        rospy.loginfo("Rotating to angle %f rad" %angle)

        while True:
            end_angle = yaw

            if end_angle < 0 + angle:
                rotateMessage.angular.z = -end_angle + angle
            else:
                rotateMessage.angular.z = end_angle + angle

            self.velocityPublisher.publish(rotateMessage)
            loop_rate.sleep()

            if (end_angle < 0.02 + angle and end_angle > -0.02 + angle):
                break

        self.velocityPublisher.publish(rotateMessage)
        loop_rate.sleep()

    def goToGoal(self, x2, y2):
        global x, y, obstacle_in_front, successful_measurement

        dx = x2 - round(x, 2)
        dy = y2 - round(y, 2)
        theta = math.atan2(dy, dx)
        distance = sqrt(dx * dx + dy * dy)
        self.rotateToAngle(theta)
        if not obstacle_in_front:
            self.move(2, distance)
        else:
            print("Not reachable! go_to_goal()")
            successful_measurement = False

    def __init__(self):

        self.turtlebot_odom_pose = Odometry()
        self.velocityPublisher = rospy.Publisher('/cmd_vel_mux/input/teleop', Twist, queue_size=10)
        rospy.Subscriber("odom", Odometry, self.odom_callback)
        # 2 HZ
        r = rospy.Rate(20)
        r.sleep()




# grid class serves either for collecting fingerprint samples
# or for very rough map
class Grid:
    g = defaultdict(dict)

    #extreme vlues of the grid
    minX, maxX, minY, maxY = 0, 0, 0, 0

    def __getitem__(self, item):
        return self.g.get(item)

    def printGrid(self):
        for x in self.g:
            print x

    def isSet(self, x, y):
        if tuple([x,y]) in self.g:
            return True


    def addPoint(self, x, y, sample):
        #extreme values check
        if x > self.maxX:
            self.maxX = x
        elif x < self.minX:
            self.mminX = x

        if y > self.maxY:
            self.maxY = y
        elif x < self.minY:
            self.mminY = y

        self.g[tuple([x, y])] = sample



def measureFP(goalX, goalY):
    global successful_measurement
    global x, y
    originX = x
    originY = y
    successful_measurement = False

    RobotMoveControl().rotateInit()
    RobotMoveControl().rotateToAngle(math.pi/2)
    RobotMoveControl().rotateToAngle(math.pi)
    RobotMoveControl().rotateToAngle(-math.pi/2)
    goal = MoveBaseGoal()
    #frame parameters
    goal.target_pose.header.frame_id = "map"
    goal.target_pose.header.stamp = rospy.Time.now()


    # navigating to goal location
    goal.target_pose.pose.position =  Point(xGoal,yGoal,0)
    goal.target_pose.pose.orientation.x = 0.0
    goal.target_pose.pose.orientation.y = 0.0
    goal.target_pose.pose.orientation.z = 0.0
    goal.target_pose.pose.orientation.w = 1.0

    action_client.send_goal(goal)
    action_client.wait_for_result(rospy.Duration(90))

    if(action_client.get_state() ==  GoalStatus.SUCCEEDED):
        rospy.loginfo("success")
        successful_measurement = True
    else:
        rospy.loginfo("Didn't reach the goal move_to_goal")
        successful_measurement = False
    if (successful_measurement):
        grid.addPoint(goalX, goalY, "measured values")
    else:
        RobotMoveControl().goToGoal(originX, originY)

#move to the goal location using move_base
def go_to_goal(xGoal,yGoal):
    action_client = actionlib.SimpleActionClient("move_base", MoveBaseAction)
    while(not action_client.wait_for_server(rospy.Duration.from_sec(5.0))):
        rospy.loginfo("Waiting for the move_base action server...")

    goal = MoveBaseGoal()
    #frame parameters
    goal.target_pose.header.frame_id = "map"
    goal.target_pose.header.stamp = rospy.Time.now()


    # navigating to goal location
    goal.target_pose.pose.position =  Point(xGoal,yGoal,0)
    goal.target_pose.pose.orientation.x = 0.0
    goal.target_pose.pose.orientation.y = 0.0
    goal.target_pose.pose.orientation.z = 0.0
    goal.target_pose.pose.orientation.w = 1.0

    action_client.send_goal(goal)
    action_client.wait_for_result(rospy.Duration(90))

    if(action_client.get_state() ==  GoalStatus.SUCCEEDED):
        rospy.loginfo("success")
        successful_measurement = True
    else:
        rospy.loginfo("Didn't reach the goal move_to_goal - move_base")
        successful_measurement = False


def collectFingerprints():
    global x, y, spacing
    gx = x
    gy = y

    #client for sending goal requests
    action_client = actionlib.SimpleActionClient("move_base", MoveBaseAction)
    while(not action_client.wait_for_server(rospy.Duration.from_sec(5.0))):
        rospy.loginfo("Waiting for the move_base action server...")

    theStack = [(gx, gy)]

    while len(theStack) > 0:

        gx, gy = theStack.pop()
        grid.printGrid()

        if (grid.isSet(gx, gy) or rawMap[(gx + spacing, gy)] == False):
            print("Coordinates already measured or unreachable.")
            continue

        successful_measurement = False

        RobotMoveControl().rotateInit()
        RobotMoveControl().rotateToAngle(math.pi/2)
        RobotMoveControl().rotateToAngle(math.pi)
        RobotMoveControl().rotateToAngle(-math.pi/2)
        goal = MoveBaseGoal()
        #frame parameters
        goal.target_pose.header.frame_id = "map"
        goal.target_pose.header.stamp = rospy.Time.now()


        # navigating to goal location
        goal.target_pose.pose.position =  Point(gx,gy,0)
        goal.target_pose.pose.orientation.x = 0.0
        goal.target_pose.pose.orientation.y = 0.0
        goal.target_pose.pose.orientation.z = 0.0
        goal.target_pose.pose.orientation.w = 1.0

        action_client.send_goal(goal)
        action_client.wait_for_result(rospy.Duration(90))

        if(action_client.get_state() ==  GoalStatus.SUCCEEDED):
            rospy.loginfo("success")
            successful_measurement = True
        else:
            rospy.loginfo("Didn't reach the goal move_to_goal")
            successful_measurement = False
        if (successful_measurement):
            grid.addPoint(gx, gy, "measured values")
        else:
            RobotMoveControl().goToGoal(gx, gy)

        theStack.append((gx + spacing, gy))  # right
        theStack.append((gx + spacing, gy + spacing))  # up right
        theStack.append((gx - spacing, gy))  # left
        theStack.append((gx - spacing, gy - spacing))  # down left
        theStack.append((gx, gy - spacing))  # down
        theStack.append((gx + spacing, gy - spacing))  # down right
        theStack.append((gx, gy + spacing))  # up
        theStack.append((gx - spacing, gy + spacing))  # up left



if __name__ == '__main__':
    try:
        rospy.init_node('UHK_robot_control')



        RobotMoveControl()
        grid = Grid()
        collectFingerprints()

        time.sleep(1.0)

    except rospy.ROSInterruptException:
        rospy.loginfo("node terminated.")

