/*
* Ultrasonic Sensor HY-SRF05
* Publishing to ROS (Robot operating system) 
* Grant permissions:
*     sudo chmod a+rw /dev/ttyUSB0
* 
* Launch the rosserial node:
*    rosrun rosserial_python serial_node.py /dev/ttyUSB0
*  
* Subscribe the topic /ultrasound_range
*  
* Ludek Krula UHK
* 
*/

#include <ros.h>
#include <ros/time.h>
#include <sensor_msgs/Range.h> 

ros::NodeHandle nh; 
sensor_msgs::Range range_msg;
ros::Publisher pub_range( "/ultrasound_range", &range_msg);
char frameid[] = "/ultrasound"; 

// pin numbers definition
const int trigPin = 9;
const int echoPin = 10;
const boolean CENTIMETERS = true;
const boolean INCHES = false; 

//variables
long duration;
long inch;
long cm;

void setup() {
  // initialize serial communication: 
  nh.initNode();
  nh.advertise(pub_range);

  //sensor parameters setup
  range_msg.radiation_type = sensor_msgs::Range::ULTRASOUND;
  range_msg.header.frame_id = frameid;
  range_msg.field_of_view = 0.1; // nearly zero (ultrasound returns just one value no FOV)
  range_msg.min_range = 0.002; // 2 cm 
  range_msg.max_range = 0.150; // 150 cm 

  pinMode(trigPin, OUTPUT); // Sets the trigPin as an Output
  pinMode(echoPin, INPUT); // Sets the echoPin as an Input
  //Serial.begin(9600); // Starts the serial communication just for debugging

}

void loop() {
  range_msg.range=getRange();
  range_msg.header.stamp = nh.now();
  pub_range.publish(&range_msg);
  nh.spinOnce();
  delay(1); 
}

long getRange() {
  // Pulsing the trigPin
  digitalWrite(trigPin, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);

  // Reading the echoPin, time the sound vawe traveled to the obstacle and back
  duration = pulseIn(echoPin, HIGH);

  // Calculating the distance
  // inch = microsecondsToInches(duration);
   cm = microsecondsToCentimeters(duration);
   // Prints the distance on the Serial Monitor
   /*Serial.print(inches);
   Serial.print("in, ");
   Serial.print(cm);
   Serial.print("cm");
   Serial.println();
    */
   return cm;
}

long microsecondsToInches(long microseconds) {
  //conversion from time of the sound vawe trip to inch
  //devided by 2 - traveled there and back to echo pin
   return microseconds / 74 / 2;
}

long microsecondsToCentimeters(long microseconds) {
  //conversion from time of the sound vawe trip to cm
  //devided by 2 - echo traveled there and back again
   return microseconds / 29 / 2;
}
